﻿using RestSharp;
using RestSharp.Authenticators;
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.IO;
using System.Linq;
using System.Net;
using System.Net.Http;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;

namespace DigitalReceiptCloudSample
{
    public partial class SampleForm : Form
    {

        private string _jwtToken;
        private string _refreshToken;
        private const string CloudUrl = "https://api.retailforce.cloud/api/v1.0/";
        private const string CloudFunctionAuthenticate = "authenticate/logon";
        private const string AuthenticateParamNameApiKey = "key";
        private const string AuthenticateParamNameApiSecret = "secret";
        private const string RefreshTokenName = "RefreshToken";
        private const string CloudFunctionLicenseToken = "licensing/token";        
        internal const string LicenseTokenParamNameClientId = "uniqueClientId";
        internal const string LicenseTokenParamNameAccessLicenseId = "accesslicenseId";
        private const string ApiKey = "RFC_32f3ced51dad4b2baffeb90a4a4f8938"; 
        private const string ApiSecret = "WIz02QhzThA_S72p4NYKMe4l_DQbWzOPNw";
        private static readonly Guid UniqueClientId = new Guid("D5ECC344-E37A-40C6-BED4-A6B1E60E68C8");
        private static readonly Guid OrganisationId = new Guid("1C8DFE2C-64CA-4114-9AFD-32907222A98B");
        private const string AccessLicenseId = "DigitalReceiptStore";        
        private const string UploadFunctionUrl = "https://functions.retailforce.cloud/api/digitalReceipt";
        private const string UploadFunctionParamNameClientId = "uniqueClientId";
        private const string UploadFunctionParamNameLicenseToken = "licenseToken";
        private const string UploadFunctionParamNameFile = "file";


        public SampleForm()
        {
            InitializeComponent();
        }

        private void btnFile_Click(object sender, EventArgs e)
        {
            OpenFileDialog fileDialog = new OpenFileDialog();
            fileDialog.Title = "Datei auswählen";
            fileDialog.InitialDirectory = @"c:\";
            fileDialog.Filter = "All files (*.*)|*.*|All files (*.*)|*.*";
            fileDialog.FilterIndex = 2;
            fileDialog.RestoreDirectory = true;
            if (fileDialog.ShowDialog() == DialogResult.OK)
            {
                txtFileName.Text = fileDialog.FileName;
            }
        }

        private void btnUpload_Click(object sender, EventArgs e)
        {
            if (string.IsNullOrEmpty(txtFileName.Text))
            {
                MessageBox.Show ("Sie müssen eine Datei auswählen!", "Fehler", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
            else
            {
               txtDownlaodUrl.Text = ProcessUpload (txtFileName.Text);
            }
        }

        private string ProcessUpload(string fileName)
        {
            if (string.IsNullOrEmpty(fileName)) throw new ArgumentNullException(nameof(fileName));

            Authenticate(ApiKey, ApiSecret);
            string token = RequestLicenseToken(UniqueClientId, AccessLicenseId);
            return UploadDigitalDocument(UniqueClientId, token, fileName);            
        }


        public void Authenticate(string cloudApiKey, string cloudApiSecret)
        {
            // parameter check
            if (string.IsNullOrEmpty(cloudApiKey)) throw new ArgumentNullException(nameof(cloudApiKey));
            if (string.IsNullOrEmpty(cloudApiSecret)) throw new ArgumentNullException(nameof(cloudApiSecret));

            // set security protocol
            ServicePointManager.SecurityProtocol |= SecurityProtocolType.Tls12;

            // it is possible to authenticate again (without signout) to avoid a block if server disconnects the user.

            RestClient client = CreateClient();
            RestRequest request = CreateRequest(CloudFunctionAuthenticate, Method.POST, 
                new Parameter(AuthenticateParamNameApiKey, cloudApiKey, ParameterType.GetOrPost), 
                new Parameter(AuthenticateParamNameApiSecret, cloudApiSecret, ParameterType.GetOrPost));

            var response = client.Post<string>(request);

            if (!response.IsSuccessful)
            {
                if (response.StatusCode == HttpStatusCode.Unauthorized)
                {
                    throw new Exception("Not possible to authenticate to cloud.");
                }
                throw new Exception(response.ErrorMessage, response.ErrorException);
            }
            
            string tmpToken = response.Cookies.FirstOrDefault(c => c.Name.Equals(RefreshTokenName))?.Value;
            if (!string.IsNullOrEmpty(tmpToken))
            {
                _refreshToken = tmpToken;
            }
            else
            {
                throw new Exception("refreshToken not found");
            }

            string jwtToken = response.Data;
            if (!string.IsNullOrEmpty(jwtToken))
                _jwtToken = jwtToken;            
        }

        /// <summary>
        /// requests a upload token from cloud
        /// </summary>
        /// <param name="uniqueClientId">Client id from the fiscalisation</param>
        /// <param name="accessLicenseId"></param>
        /// <returns>The license token</returns>
        public string RequestLicenseToken(Guid uniqueClientId, string accessLicenseId)
        {
            if (uniqueClientId == null || uniqueClientId == Guid.Empty) throw new ArgumentNullException(nameof(uniqueClientId));
            string result;

            try
            {
                RestClient client = CreateClient();
                RestRequest request = CreateRequest(CloudFunctionLicenseToken, Method.GET, 
                new Parameter(LicenseTokenParamNameClientId, uniqueClientId, ParameterType.QueryString), 
                new Parameter(LicenseTokenParamNameAccessLicenseId, accessLicenseId, ParameterType.QueryString));

                IRestResponse<string> response = client.Get<string>(request);
                if (response.StatusCode == HttpStatusCode.OK)
                {
                    return response.Data;
                }
                throw new Exception($"Request failed ({response.ErrorMessage})!");                
            }
            catch (Exception e)
            {
                throw new Exception("Error can't get token!", e);
            }            
        }


        /// <summary>
        /// uploads a file and generates a downloadurl per function (attention expiration have to be defined in sas function)
        /// </summary>
        /// <param name="uniqueClientId"></param>
        /// <param name="licenseToken"></param>
        /// <param name="filePath"></param>
        /// <returns>sas url to download the file</returns>
        public string UploadDigitalDocument(Guid uniqueClientId, string licenseToken, string filePath)
        {
            if (string.IsNullOrEmpty(licenseToken)) throw new ArgumentNullException(nameof(licenseToken));
            if (uniqueClientId == null || uniqueClientId == Guid.Empty) throw new ArgumentNullException(nameof(uniqueClientId));
            if (string.IsNullOrEmpty(filePath)) throw new ArgumentNullException(nameof(filePath));
            if (!File.Exists(filePath)) throw new ArgumentOutOfRangeException("File doesn't exists.");

            HttpClient client = new HttpClient();
            MultipartFormDataContent form = new MultipartFormDataContent();

            form.Add(new StringContent(uniqueClientId.ToString()), UploadFunctionParamNameClientId);
            form.Add(new StringContent(licenseToken), UploadFunctionParamNameLicenseToken);

            using (FileStream stream = new FileStream(filePath, FileMode.Open))

            {
                form.Add(new StreamContent(stream), UploadFunctionParamNameFile, Path.GetFileName(filePath));
                HttpResponseMessage response = null;
                try
                {
                    response = (client.PostAsync(UploadFunctionUrl, form)).Result;
                    if (response.IsSuccessStatusCode)
                    {
                        var k = response.Content.ReadAsStringAsync().Result;
                        return k.ToString();
                    }  
                }
                catch (Exception ex)
                {
                    Console.WriteLine(ex.Message);
                }
                return null;
            }
        }

        private RestClient CreateClient()
        {
            RestClient restClient = new RestClient(CloudUrl);
            if (!string.IsNullOrEmpty(_jwtToken))
                restClient.Authenticator = new JwtAuthenticator(_jwtToken);

            return restClient;
        }

        private RestRequest CreateRequest(string url, Method method, params Parameter[] parameters)
        {
            var request = new RestRequest(url, method, DataFormat.Json);

            if (parameters != null && parameters.Length > 0)
                request.Parameters.AddRange(parameters);

            return request;
        }
    }
}
